from ipywidgets import HTML
from traitlets import Unicode
from IPython.display import display
import ipyvuetify as v 

# change vuetify theming 
v.theme.dark = True
v.theme.themes.dark.primary = "#B3842E"
v.theme.themes.dark.accent = "#a1458e"
v.theme.themes.dark.secondary = "#324a88"
v.theme.themes.dark.success = "#3F802A"
v.theme.themes.dark.info = "#79B1C9"
v.theme.themes.dark.warning = "#b8721d"
v.theme.themes.dark.error = "#A63228"

#fixed colors 
sepal_main = '#24221F'
sepal_darker = '#1a1a1a'


# Fixed styles to avoid leaflet maps overlap sepal widgets

class Styles(v.VuetifyTemplate):

    template=Unicode("""
    <style>
        .leaflet-pane {
            z-index : 2 !important;
        }
        .leaflet-top, .leaflet-bottom {
            z-index : 2 !important;
        }
        main.v-content {
            padding-top: 0px !important;
        }
    </style>
    """).tag(sync=True)
    
styles=Styles()
display(styles)

COMPONENTS = {

    'PROGRESS_BAR':{
        'color':'indigo',
    }
}

ICON_TYPES = {
    # Used for folders
    '':{ 
        'color':'amber',
        'icon':'mdi-folder-outline'
    },
    '.csv':{
        'color':'green accent-4',
        'icon':'mdi-border-all'
    },
    '.txt':{
        'color':'green accent-4',
        'icon':'mdi-border-all'
    },
    '.tif':{
        'color':'deep-purple',
        'icon':'mdi-image-outline'
    },
    '.tiff':{
        'color':'deep-purple',
        'icon':'mdi-image-outline'
    },
    '.shp':{
        'color':'deep-purple',
        'icon':'mdi-vector-polyline'
    },
    'DEFAULT':{
        'color':'light-blue',
        'icon':'mdi-file-outline'
    },
    # Icon for parent folder
    'PARENT':{ 
        'color':'white',
        'icon':'mdi-folder-upload-outline'
    },

}



import time
import sys
import re
import json
from pathlib import Path

import ee
import geemap
import pandas as pd
import geopandas as gpd

from sepal_ui.scripts import utils, gee
from sepal_ui.message import ms
from sepal_ui.scripts import utils as su
    
# initialize earth engine
su.init_ee()

def display_asset(output, asset):
    """
    remove the manifest from the asset name and display it to the user in an output
    
    Args: 
        output (sw.Alert): the alert to display information to the user
        asset (str): the asset name
    
    Return:
        (str): the asset name without the manifest
    """
    
    asset = asset.replace('projects/earthengine-legacy/assets/', '')
    
    output.add_msg(ms.aoi_sel.asset_created.format(asset), 'success')
    
    return asset

def isAsset(asset_descripsion, folder):
    """
    Check if the asset already exist in the user asset folder
    
    Args: 
        asset_descripsion (str) : the descripsion of the asset
        folder (str): the folder of the glad assets
        
    Return:
        (bool): true if already in folder
    """
    exist = False
    liste = ee.data.listAssets({'parent': folder})['assets']
    for asset in liste:
        if asset['name'] == str(Path(folder,asset_descripsion)):
            exist = True
            break
            
    return exist 

def get_admin0_asset(adm0, output):
    """
    send a request to GEE to get a featurecollection based on the country selected
    
    Args:
        adm0 (int): the administrative level 0 in FAO GAUL 2015
        output (sw.Alert): the alert to display information to the end user
        
    return:
        (ee.FeatureColection): the FeatureCollection invocation of the country
    """
    
    if adm0 == None:
        output.add_msg(ms.aoi_sel.no_country, 'warning')
        return (None, None)
    
    country = ee.FeatureCollection("FAO/GAUL/2015/level0").filter(ee.Filter.eq('ADM0_CODE', adm0))
    
    #output.add_msg(ms.aoi_sel.valid_admin0.format(iso_3), 'success')
    
    return country

def get_admin1_asset(adm1, output):
    """
    send a request to GEE to get a featurecollection based on the country selected
    
    Args:
        adm1 (int): the administrative level 1 in FAO GAUL 2015
        output (sw.Alert): the alert to display information to the end user
        
    return:
        (ee.FeatureColection): the FeatureCollection invocation of the country
    """
    
    if adm1 == None:
        output.add_msg(ms.aoi_sel.no_country, 'warning')
        return (None, None)
    
    country = ee.FeatureCollection("FAO/GAUL/2015/level1").filter(ee.Filter.eq('ADM1_CODE', adm1))
    
    #
    
    return country

def get_admin2_asset(adm2, output):
    """
    send a request to GEE to get a featurecollection based on the country selected
    
    Args:
        adm2 (int): the administrative level 2 in FAO GAUL 2015
        output (sw.Alert): the alert to display information to the end user
        
    return:
        (ee.FeatureColection): the FeatureCollection invocation of the country
    """
    
    if adm2 == None:
        output.add_msg(ms.aoi_sel.no_country, 'warning')
        return (None, None)
    
    country = ee.FeatureCollection("FAO/GAUL/2015/level2").filter(ee.Filter.eq('ADM2_CODE', adm2))
    
    #output.add_msg(ms.aoi_sel.valid_admin0.format(iso_3), 'success')
    
    return country

def get_drawn_shape(drawn_feat, file_name, folder, output):
    """
    send the requested drawn feature to GEE to create an asset
    
    Args:
        drawn_feat (ee.FeatureCollection): the feature drawn on the map
        file_name (str): the name to use as GEE asset
        folder (str): the path to the user GEE folder
        output (sw.Alert): the alert to display information to the end user
        
    Return:
        (str): the created asset name
    """
    
    aoi = drawn_feat 
    asset_name = ms.aoi_sel.file_pattern.format(re.sub('[^a-zA-Z\d\-\_]', '_', file_name))
        
    if drawn_feat == None: #check if something is drawn 
        output.add_msg(ms.aoi_sel.no_shape, 'error')
        return None
    elif isAsset(asset_name, folder): #check asset existence
        output.add_msg(ms.aoi_sel.name_used, 'error')
        return None 
        
    asset = str(Path(folder, asset_name))
            
    # create and launch the task
    task_config = {
        'collection': drawn_feat, 
        'description':asset_name,
        'assetId': asset
    }
    task = ee.batch.Export.table.toAsset(**task_config)
    task.start()
    gee.wait_for_completion(asset_name, output)
           
    display_asset(output, asset)
    
    return asset

def get_gee_asset(asset_name, output):
    """
    check that the asset name is not null or None
    
    Args:
        asset_name (str): the asset name
        output (sw.Alert): the alert to display information to the end user
        
    Return:
        (str): return the asset if not null
    """
    
    if asset_name == '' or asset_name == None:
        output.add_msg(ms.aoi_sel.no_asset, 'error') 
        asset = None
    else:
        output.add_msg(ms.aoi_sel.check_if_asset, 'success')
        asset = asset_name
        
    return asset

def get_csv_asset(json_csv, file_name, folder, output):
    """
    Send a request to GEE to create an asset based on the local shapefile
    
    Args:
        json_csv (str): the csv file information stored as a json dict {pathname, id, lng, lat}
        file_name (str): the name that will be used in GEE 
        folder (str): The user folder in GEE
        outpput (sw.Alert): the alert to display information to the end user
        
    Return:
        (str): the created asset name
    """
    
    # read the json 
    load_df = json.loads(json_csv)
    
    # check that the columns are well set 
    tmp_list = [load_df[i] for i in load_df]
    if not len(tmp_list) == len(set(tmp_list)):
        output.add_msg(ms.aoi_sel.duplicate_key, 'error') 
        return None
    
    # check asset existence
    asset_name = ms.aoi_sel.file_pattern.format(re.sub('[^a-zA-Z\d\-\_]', '_', file_name))
    if isAsset(asset_name, folder):
        output.add_msg(ms.aoi_sel.name_used, 'error')
        return None
    
    # create a tmp gdf
    df = pd.read_csv(load_df['pathname'], sep=None, engine='python')
    gdf = gpd.GeoDataFrame(df, crs='EPSG:4326', geometry = gpd.points_from_xy(df[load_df['lng_column']], df[load_df['lat_column']]))
    
    # convert it into geo-json 
    json_df = json.loads(gdf.to_json())
    
    # create a gee object with geemap
    output.add_msg(ms.aoi_sel.geojson_to_ee, 'success')
    ee_object = geemap.geojson_to_ee(json_df)
    
    # upload this object to earthengine
    asset = str(Path(folder, asset_name))
            
    # create and launch the task
    task_config = {
        'collection': ee_object, 
        'description':asset_name,
        'assetId': asset
    }
    task = ee.batch.Export.table.toAsset(**task_config)
    task.start()
    gee.wait_for_completion(asset_name, output)
           
    display_asset(output, asset)
    
    return asset
    
            
def get_shp_aoi(file_input, file_name, folder, output):
    """
    Send a request to GEE to create an asset based on the local shapefile
    
    Args:
        file_input (str | pathlib.path): the path to the .shp file
        file_name (str): the name used for the GEE asset
        folder (str): the user folder in GEE
        output (sw.Alert): the alert to display information to the end user
        
    Return:
        (str): the created asset name
    """
    
    if type(file_input) == str:
        file_input = Path(file_input)
        
    if not file_input.is_file():
        output.add_msg(ms.aoi_sel.error_occured, 'error')
        return None
    
    # convert the .shp in gdf in epsg:4326
    gdf = gpd.read_file(file_input).to_crs('EPSG:4326')
    
    # convert gdf to geo json 
    json_gdf = json.loads(gdf.to_json())
    
    # create ee_object based on the json dict
    ee_object = geemap.geojson_to_ee(json_gdf)           
       
    asset_name = ms.aoi_sel.file_pattern.format(re.sub('[^a-zA-Z\d\-\_]','_',file_name))
        
    # check asset's name
    if isAsset(asset_name, folder):
        output.add_msg(ms.aoi_sel.name_used, 'error')
        asset = None
        
    else:
        
        asset = str(Path(folder, asset_name))
            
        # launch the task
        task_config = {
            'collection': ee_object, 
            'description':asset_name,
            'assetId': asset
        }
        task = ee.batch.Export.table.toAsset(**task_config)
        task.start()
        gee.wait_for_completion(asset_name, output)
                   
        display_asset(output, asset)
    
    return asset
    
def run_aoi_selection(output, list_method, io, folder=None):
    """
    Create an gee asset according to the user inputs

    Args:
        io.file_input (str): the file to retreive from the user folder. It must be a .shp file
        io.file_name (str): name of the aoi that will be used troughout the process
        io.country_selection (str): a country name in english available in LSIB 2017
        io.asset_name (str): the assetId of a existing asset
        io.drawing_method (str): the name of the method selected to create the asset
        output (v.Alert): the widget used to display the process informations
        list_method ([str]): list of the method use to select an AOI
        drawn_feat (ee.FeatureCollection): the last drawn object on the map
        folder (str, optional): the user GEE asset folder
        
    Return:
        (str) : the AssetId of the AOI
    """
    # go to the glad folder in gee assets 
    if not folder: 
        folder = ee.data.getAssetRoots()[0]['id']
    
    # clean all but the selected method
    if io.adm0:
        io.assetId = None
    else:
        io.feature_collection, io.country_code = (None, None)
    
    
    # check the aoi method used
    
    # not selected
    if io.selection_method == None:
        raise Exception(ms.aoi_sel.no_selection)   
    # use a country boundary
    elif io.selection_method == list_method[0]: 
        asset = get_admin0_asset(io.adm0, output)
        io.set_admin(asset, admin0=io.adm0)
        output.add_msg(ms.aoi_sel.valid_admin0.format(io.get_aoi_name()), 'success')
    # use adm1
    elif io.selection_method == list_method[1]: 
        asset = get_admin1_asset(io.adm1, output)
        io.set_admin(asset, admin0=io.adm0, admin1=io.adm1)
        output.add_msg(ms.aoi_sel.valid_admin0.format(io.get_aoi_name()), 'success')
    # use adm2
    elif io.selection_method == list_method[2]: 
        asset = get_admin2_asset(io.adm2, output)
        io.set_admin(asset, admin0=io.adm0, admin1=io.adm1, admin2=io.adm2)
        output.add_msg(ms.aoi_sel.valid_admin0.format(io.get_aoi_name()), 'success')
    # draw a shape
    elif io.selection_method == list_method[3]: 
        io.set_asset(get_drawn_shape(io.drawn_feat, io.file_name, folder, output))
    # use GEE asset
    elif io.selection_method == list_method[4]: 
        io.set_asset(get_gee_asset(io.assetId, output))
    # upload file
    elif io.selection_method == list_method[5]: 
        io.set_asset(get_shp_aoi(io.file_input, io.file_name, folder, output))
    # csv point file
    elif io.selection_method == list_method[6]: 
        io.set_asset(get_csv_asset(io.json_csv, io.file_name, folder, output))
            
    return


import ipyvuetify as v
from markdown import markdown
from traitlets import Unicode

TYPES = ('info', 'secondary', 'primary', 'error', 'warning', 'success', 'accent')

class SepalWidget(v.VuetifyWidget):
    """
    Custom vuetifyWidget to add specific methods
    
    Attributes:
        viz (bool): weather the file is displayed or not
        old_class (str): a saving attribute of the widget class
    """
    
    def __init__(self, **kwargs):
        
        super().__init__(**kwargs)
        
        self.viz = True
        self.old_class = ''
        
    def toggle_viz(self):
        """
        toogle the visibility of the widget
        
        Return:
            self
        """
        
        return self.hide() if self.viz else self.show()
    
    def hide(self):
        """
        Hide the widget by adding the d-none html class to the widget.
        Save the previous class and set viz attribute to False.
        
        Return:
            self
        """
        
        if not 'd-none' in str(self.class_):
            self.old_class = self.class_
            self.class_ = 'd-none'
            
        self.viz = False
        
        return self
        
    def show(self):
        """
        Hide the widget by removing the d-none html class to the widget
        Save the previous class and set viz attribute to True.
        
        Return:
            self
        """
        
        if self.old_class:
            self.class_ = self.old_class
        
        if 'd-none' in str(self.class_):
            self.class_ = str(self.class_).replace('d-none', '')
                
        self.viz = True
        
        return self
    
class Markdown(v.Layout, SepalWidget):
    """
    Custom Layout based on the markdown text given
    
    Args:
        mkd_str (str): the text to display using the markdown convention. multi-line string are also interpreted
    """
    
    def __init__(self, mkd_str="", **kwargs):
        
        mkd = markdown(mkd_str, extensions=['fenced_code','sane_lists'])
    
        #need to be nested in a div to be displayed
        mkd = '<div>\n' + mkd + '\n</div>'
    
        #create a Html widget
        class MyHTML(v.VuetifyTemplate):
            template = Unicode(mkd).tag(sync=True)
    
        content = MyHTML()
        
        super().__init__(
            row=True,
            class_='pa-5',
            align_center=True,
            children=[v.Flex(xs12=True, children=[content])],
            **kwargs
        )
        
class Tooltip(v.Tooltip):
    """
    Custom widget to display tooltip when mouse is over widget

    Args:
        widget (DOM.widget): widget used to display tooltip
        tooltip (str): the text to display in the tooltip            
    """
    
    def __init__(self, widget, tooltip, *args, **kwargs):
        
        
        self.bottom=True
        self.v_slots=[{
            'name': 'activator',
            'variable': 'tooltip',
            'children': widget
        }]
        widget.v_on = 'tooltip.on'
        
        self.children = [tooltip]
        
        super().__init__(*args, **kwargs)
        
    def __setattr__(self, name, value):
        """prevent set attributes after instantiate tooltip class"""
        
        if hasattr(self,'_model_id'):
            if self._model_id:
                raise RuntimeError(f"You can't modify the attributes of the {self.__class__} after instantiated")
        super().__setattr__(name, value)


from pathlib import Path
import json

import ipyvuetify as v
from traitlets import (
    HasTraits, Unicode, link, Int
)
from ipywidgets import jslink
import pandas as pd
import ee

from sepal_ui.frontend.styles import *
from sepal_ui.scripts import utils as su
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.sepalwidgets.btn import Btn

# initialize earth engine
su.init_ee()

class DatePicker(v.Layout, SepalWidget):
    """
    Custom input widget to provide a reusable DatePicker. It allows to choose date as a string in the following format YYYY-MM-DD
    
    Args:
        label (str, optional): the label of the datepicker field
        
    Attributes:
        menu (v.Menu): the menu widget to display the datepicker
        
    """
    def __init__(self, label="Date", **kwargs):
        
        date_picker = v.DatePicker(
            no_title = True, 
            v_model = None, 
            scrollable = True
        )

        date_text = v.TextField(
            v_model = None,
            label = label,
            hint = "YYYY-MM-DD format",
            persistent_hint = True, 
            prepend_icon = "event",
            readonly = True,
            v_on = 'menuData.on'
        )

        self.menu = v.Menu(
            min_width="290px",
            transition = "scale-transition",
            offset_y = True, 
            value = False,
            close_on_content_click = False,
            children = [date_picker],
            v_slots = [{
                'name': 'activator',
                'variable': 'menuData',
                'children': date_text,
            }]
        )

        super().__init__(
            v_model = None,
            row = True,
            class_ = 'pa-5',
            align_center = True,
            children = [v.Flex(xs10=True, children=[self.menu])],
            **kwargs
        )

        jslink((date_picker, 'v_model'), (date_text, 'v_model'))
        jslink((date_picker, 'v_model'), (self, 'v_model'))
        
        #close the datepicker on click
        #date_text.observe(lambda _: setattr(self.menu, 'value', False), 'v_model')
        
class FileInput(v.Flex, SepalWidget, HasTraits):
    """
    Custom input field to select a file in the sepal folders. 
    
    Args:
        extentions ([str]): the list of the allowed extentions. the FileInput will only display thiese extention and folders
        folder (str | pathlib.Path): the starting folder of the file input
        label (str): the label of the input
        v_model (str, optional): the default value
        
    Attributes:
        extentions ([str]): the extention list
        folder (str | pathlib.Path): the current folder
        selected_file (v.TextField): the textfield where the file pathname is stored
        loading (v.ProgressLinear): loading top bar of the menu component
        file_list (v.List): the list of files and folder that are available in the current folder
        file_menu (v.Menu): the menu that hide and show the file_list
        reload (v.Btn): reload btn to reload the file list on the current folder 
    """

    file = Unicode('')
    
    def __init__(self, extentions = [], folder=Path('~').expanduser(), label='search file', v_model = None, **kwargs):
        
        if type(folder) == str:
            folder = Path(folder)
            
        self.extentions = extentions
        self.folder = folder
        
        self.selected_file = v.TextField(
            readonly = True,
            label    = 'Selected file', 
            class_   = 'ml-5 mt-5',
            v_model  = self.file
        )

        self.loading = v.ProgressLinear(
            indeterminate    = False, 
            background_color = 'grey darken-3',
            color            = COMPONENTS['PROGRESS_BAR']['color']
            )
        
        self.file_list = v.List(
            dense      = True, 
            color      = 'grey darken-3',
            flat       = True,
            max_height = '300px',
            style_     = 'overflow: auto; border-radius: 0 0 0 0;',
            children   = [ 
                v.ListItemGroup(
                    children = self._get_items(),
                    v_model  = ''
                )
            ]
        )

        self.file_menu = v.Menu(
            min_width              = 300,
            children               = [self.loading, self.file_list], 
            close_on_content_click = False,
            v_slots                = [{
                'name': 'activator',
                'variable': 'x',
                'children': Btn(icon='mdi-file-search', v_model=False, v_on='x.on', text=label)
        }])
        
        self.reload = v.Btn(
            icon = True,
            color = 'primary',
            children = [v.Icon(children=['mdi-cached'])]
        )
        
        super().__init__(
            row          = True,
            class_       = 'd-flex align-center mb-2',
            align_center = True,
            children     = [
                self.reload,
                self.file_menu,
                self.selected_file,
            ],
            **kwargs
        )
        
        link((self.selected_file, 'v_model'), (self, 'file'))
        link((self.selected_file, 'v_model'), (self, 'v_model'))

        self.file_list.children[0].observe(self._on_file_select, 'v_model')
        self.reload.on_event('click', self._on_reload)
        
    def _on_file_select(self, change):
        """Dispatch the behaviour between file selection and folder change"""
        
        if not change['new']:
            return self
        
        new_value = Path(change['new'])
        
        if new_value.is_dir():
            self.folder = new_value
            self._change_folder()
                
        elif new_value.is_file():
            self.file = str(new_value)
            
        return self
                
    def _change_folder(self):
        """Change the target folder"""
        #reset files
        self.file_list.children[0].children = self._get_items()
        
        return
    
    def _get_items(self):
        """Return the list of items inside the folder"""

        self.loading.indeterminate = True
        
        folder = self.folder

        list_dir = [el for el in folder.glob('*/') if not el.name.startswith('.')]

        if self.extentions:
            list_dir = [el for el in list_dir if el.is_dir() or el.suffix in self.extentions]

        folder_list = []
        file_list = []

        for el in list_dir:
            
            if el.suffix in ICON_TYPES.keys():
                icon = ICON_TYPES[el.suffix]['icon']
                color = ICON_TYPES[el.suffix]['color']
            else:
                icon = ICON_TYPES['DEFAULT']['icon']
                color = ICON_TYPES['DEFAULT']['color']
            
            children = [
                v.ListItemAction(children=[v.Icon(color= color,children=[icon])]),
                v.ListItemContent(children=[v.ListItemTitle(children=[el.stem + el.suffix])]),
            ] 

            if el.is_dir():
                folder_list.append(v.ListItem(value=str(el), children=children))
            else:
                file_size = su.get_file_size(el)
                children.append(v.ListItemActionText(children=[file_size]))
                file_list.append(v.ListItem(value=str(el), children=children))

        folder_list = sorted(folder_list, key=lambda x: x.value)
        file_list = sorted(file_list, key=lambda x: x.value)

        parent_item = v.ListItem(
            value=str(folder.parent), 
            children=[
                v.ListItemAction(children=[v.Icon(color=ICON_TYPES['PARENT']['color'], children=[ICON_TYPES['PARENT']['icon']])]),
                v.ListItemContent(children=[v.ListItemTitle(children=[f'..{folder.parent}'])]),
            ]
        )

        folder_list.extend(file_list)
        folder_list.insert(0,parent_item)

        self.loading.indeterminate = False
        
        return folder_list
    
    def _on_reload(self, widget, event, data):
        
        # force the update of the current folder
        self._change_folder()
        
        return

class LoadTableField(v.Col, SepalWidget):
    """
    A custom input widget to load points data. The user will provide a csv or txt file containing labeled dataset. 
    The relevant columns (lat, long and id) can then be identified in the updated select. Once everything is set, the widget will populate itself with a json dict.
    {pathname, id_column, lat_column,lng_column}
    
    Attributes:
        fileInput (sw.FileInput): the file input to select the .csv or .txt file
        IdSelect (v.Select): input to select the id column
        LngSelect (v.Select): input to select the lng column
        LatSelect (v.Select): input to select the lat column
    """
    
    default_v_model = {
            'pathname'  : None, 
            'id_column' : None, 
            'lat_column': None, 
            'lng_column': None
    }
    
    def __init__(self):
        
        self.fileInput = FileInput(['.csv', '.txt'])
        
        self.IdSelect = v.Select(
            _metadata = {'name': 'id_column'}, 
            items     = [], 
            label     = 'Id', 
            v_model   = None
        ) 
        self.LngSelect = v.Select(
            _metadata = {'name': 'lng_column'}, 
            items     = [], 
            label     = 'Longitude', 
            v_model   = None
        )
        self.LatSelect = v.Select(
            _metadata = {'name': 'lat_column'}, 
            items     = [], 
            label     = 'Latitude', 
            v_model   = None
        )
        
        super().__init__(
            v_model = json.dumps(self.default_v_model),
            children = [
                self.fileInput,
                self.IdSelect,
                self.LngSelect,
                self.LatSelect
            ]
        )
        
        # link the dropdowns
        jslink((self.IdSelect, 'items'),(self.LngSelect, 'items'))
        jslink((self.IdSelect, 'items'),(self.LatSelect, 'items'))
        
        # link the widget with v_model 
        self.fileInput.observe(self._on_file_input_change, 'v_model')
        self.IdSelect.observe(self._on_select_change, 'v_model')
        self.LngSelect.observe(self._on_select_change, 'v_model')
        self.LatSelect.observe(self._on_select_change, 'v_model')
        
    def _on_file_input_change(self, change):
        """Update the select content when the fileinput v_model is changing"""
        path = change['new']
            
        df = pd.read_csv(path, sep=None, engine='python')
        
        if len(df.columns) < 3: 
            self._clear_select()
            return 
        
        self._set_value('pathname', path)
        
        # clear the selects
        self._clear_select()
        self.IdSelect.items = df.columns.tolist()
        
        # pre load values that sounds like what we are looking for 
        # it will only keep the first occurence of each one 
        for name in reversed(df.columns.tolist()):
            lname = name.lower() 
            if 'id' in lname:
                self.IdSelect.v_model = name
            elif any(ext in lname for ext in ['lng', 'long', 'longitude', 'x_coord', 'xcoord']):
                self.LngSelect.v_model = name
            elif any(ext in lname for ext in ['lat', 'latitude', 'y_coord', 'ycoord']):
                self.LatSelect.v_model = name
        
        return self
                
    def _clear_select(self):
        """clear the selects components"""
        
        self.IdSelect.items = [] # all the others are listening to this one 
        self.IdSelect.v_model = self.LngSelect.v_model = self.LatSelect.v_model = None
        
        return self
    
    def _on_select_change(self, change):
        """change the v_model value when a select is changed"""
        
        name = change['owner']._metadata['name']
        self._set_value(name, change['new'])
        
        return self
        
    def _set_value(self, name, value):
        """ set the value in the json dictionary"""
        
        tmp = json.loads(self.v_model)
        tmp[name] = value
        self.v_model = json.dumps(tmp)
        
        return self
    
    def get_v_model(self):
        """
        Return the v_model as a dict
        
        Return:
            (dict): the v_model
        """
        
        return json.loads(self.v_model)
    
    def get_pathname(self):
        """
        Return the pathname from v_model
        
        Return:
            (str): the v_model pathname
        """
        
        return json.loads(self.v_model)['pathname']
    
    def get_id_lbl(self):
        """
        Return the id column label from v_model
        
        Return:
            (str): label of the id column
        """
        
        return json.loads(self.v_model)['id_column']
    
    def get_lng_lbl(self):
        """
        Return the longitude column label from v_model
        
        Return:
            (str): label of the longitude column
        """
        
        return json.loads(self.v_model)['lng_column']
    
    def get_lat_lbl(self):
        """
        Return the latitude column label from v_model
        
        Return:
            (str); label of the latitude column
        """
        
        return json.loads(self.v_model)['lat_column']

class AssetSelect(v.Combobox, SepalWidget):
    """
    Custom widget input to select an asset inside the asset folder of the user
    
    Args:
        label (str): the label of the input
        folder (str): the folder of the user assets
        default_asset (str): the id of a default asset
        
    Attributes:
        folder (str): the folder of the user assets
    """
    
    def __init__(self, label = 'Select an asset', folder = None, default_asset = None):
        
        # if folder is not set use the root one 
        self.folder = folder if folder else ee.data.getAssetRoots()[0]['id'] + '/'
        
        # get the list of user asset
        assets = ee.data.listAssets({'parent': self.folder})['assets']
        
        items = [asset['id'] for asset in assets]
        
        super().__init__(
            clearable       = True,
            class_          = 'mb-5',
            label           = label,
            placeholder     = 'users/someCustomUser/customAsset',
            hint            = "select an asset in the list or write a custom asset name. Be careful that you need to have access to this asset to use it",
            persistent_hint = True,
            items           = items,
            v_model         = default_asset
        )
        
class PasswordField(v.TextField, SepalWidget):
    """
    Custom widget to input passwords in text area and 
    toggle its visibility.

    Args:
        label (str, optional): Header displayed in text area. Defaults to Password.
    """
    def __init__(self, **kwargs):
        
        # default behaviour 
        self.label="Password"
        self.class_='mr-2'
        self.v_model=''
        self.type='password'
        self.append_icon='mdi-eye-off'
        
        # init the widget with the remaining kwargs
        super().__init__(**kwargs)
        
        # bind the js behaviour
        self.on_event('click:append' ,self._toggle_pwd)
    

    def _toggle_pwd(self, widget, event, data):
        """Toggle password visibility when append button is clicked"""
        
        if self.type=='text':
            self.type='password'
            self.append_icon = 'mdi-eye-off'
        else:
            self.type = 'text'
            self.append_icon = 'mdi-eye'
            
class NumberField(v.TextField, SepalWidget):
    """
    Custom widget to input numbers in text area and add/substract with single increment.

    Args:
        max_ (int, optional): Maximum selectable number. Defaults to 10.
        min_ (int, optional): Minimum selectable number. Defaults to 0.

    """
    max_ = Int(10).tag(sync=True)
    min_ = Int(0).tag(sync=True)
    
    def __init__(self, **kwargs):
        
        self.type='number'
        self.append_outer_icon='mdi-plus'
        self.prepend_icon='mdi-minus'
        self.v_model=0
        self.readonly=True
        
        super().__init__(**kwargs)
        
        self.on_event('click:append-outer', self.increment)
        self.on_event('click:prepend', self.decrement)
    
    def increment(self, widget, event, data):
        """Adds 1 to the current v_model number"""
        if self.v_model < self.max_: self.v_model+=1
        
    def decrement(self, widget, event, data):
        """Substracts 1 to the current v_model number"""
        if self.v_model > self.min_: self.v_model-=1



import ipyvuetify as v 
from traitlets import Int, Unicode
from IPython.display import display

class ResizeTrigger(v.VuetifyTemplate):
    template = Unicode('''
    <script>
        modules.export = {
            watch: {
                resize() {
                    window.dispatchEvent(new Event('resize'));
                }
            }
        }
        </script>
    ''').tag(sync=True)
    
    resize = Int(0).tag(sync=True)

rt = ResizeTrigger()
display(rt)



from functools import partial
from datetime import datetime

import ipyvuetify as v
import traitlets

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.frontend.styles import *
from sepal_ui.frontend.js import *

class AppBar (v.AppBar, SepalWidget):
    """
    Custom AppBar widget with the provided title using the sepal color framework
    
    Args: 
        title (str, optional): the title of the app
        
    Attributes:
        toggle_button (v.Btn): the btn to display or hide the drawer to the user
        title (v.ToolBarTitle): the widget containing the app title
    """
    
    def __init__(self, title='SEPAL module', **kwargs):
        
        self.toggle_button = v.Btn(
            icon = True, 
            children=[
                v.Icon(class_="white--text", children=['mdi-dots-vertical'])
            ]
        )
        
        self.title = v.ToolbarTitle(children=[title])
        
        super().__init__(
            color=sepal_main,
            class_="white--text",
            dense=True,
            app = True,
            children = [self.toggle_button, self.title],
            **kwargs
        )
        
    def set_title(self, title):
        """
        Set the title of the appBar
        
        Args:
            title (str): the new app title
            
        Return:
            self
        """
        
        self.title.children = [title]
            
        return self
            
class DrawerItem(v.ListItem, SepalWidget):
    """
    Custom DrawerItem using the user input. 
    If a card is set the drawerItem will trigger the display of all the Tiles in the app that have the same mount_id.
    If an href is set, the drawer will open the link in a new tab
    
    Args:
        title (str): the title of the drawer item
        icon(str, optional): the full name of a mdi-icon
        card (str, optional): the mount_id of tiles in the app
        href (str, optional): the absolute link to an external webpage
        
    Attributes:
        href (str): the absolute link to follow on click 
        _metadata (dict): single key dict to point to a mount_id. This ount_id will be used as parameter to know which tiles are to be shown on click event
        rt (ResizeTrigger): the trigger to resize maps and other javascript object when jumping from a tile to another
    """
    
    def __init__(self, title, icon=None, card=None, href=None, **kwargs):
        
        icon = icon if icon else 'mdi-folder-outline'
        
        children = [
            v.ListItemAction(
                children=[
                    v.Icon(
                        class_="white--text", 
                        children=[icon])
                ]
            ),
            v.ListItemContent(
                children=[
                    v.ListItemTitle(
                        class_="white--text", 
                        children=[title]
                    )
                ]
            )
        ]
        
        super().__init__(
            link=True,
            children=children,
            **kwargs) 

        if href:
            self.href=href
            self.target="_blank"
        elif card:
            self._metadata = {'card_id': card }
        
                        
    def display_tile(self, tiles):
        """
        Display the apropriate tiles when the item is clicked.
        The tile to display will be all tile in the list with the mount_id as the current object
    
        Args:
            tiles ([sw.Tile]) : the list of all the available tiles in the app
            
        Return:
            self
        """            
    
        self.on_event('click', partial(self._on_click, tiles=tiles))
        
        return self
    
    def _on_click(self, widget, event, data, tiles):
        
        for tile in tiles:
            if self._metadata['card_id'] == tile._metadata['mount_id']:
                tile.show()
            else:
                tile.hide()

        # trigger the risize event 
        rt.resize += 1
                
        # change the cuurent item status 
        self.input_value = True
        
        return self
            
class NavDrawer(v.NavigationDrawer, SepalWidget):
    """ 
    Custom NavDrawer using the different DrawerItems of the user and the sepal color framework. 
    The drawer can include links to the github page of the project for wiki, bugs and repository.
    
    Args:
        item ([sw.DrawerItem]): the list of all the drawerItem to display in the drawer. This items should pilote the different tile visibility
        code (str, optional): the absolute link to the source code 
        wiki (str, optional): the absolute link the the wiki page
        issue (str, optional): the absolute link to the issue tracker
    """
        
    def __init__(self, items=[], code=None, wiki=None, issue=None, **kwargs):
        
        self.items = items
        
        code_link = []
        if code:
            item_code = DrawerItem('Source code', icon='mdi-file-code', href=code)
            code_link.append(item_code)
        if wiki:
            item_wiki = DrawerItem('Wiki', icon='mdi-book-open-page-variant', href=wiki)
            code_link.append(item_wiki)
        if issue:
            item_bug = DrawerItem('Bug report', icon='mdi-bug', href=issue)
            code_link.append(item_bug)
            
        super().__init__(
            v_model=True,
            app=True,
            color = sepal_darker,
            children = [
                v.List(dense=True, children=self.items),
                v.Divider(),
                v.List(dense=True, children=code_link)
            ],
            **kwargs
        )
        
        # bind the javascripts behaviour
        for i in self.items:
            i.observe(self._on_item_click, 'input_value')
        
    def display_drawer(self, toggleButton):
        """
        Bind the drawer to the app toggleButton

        Args:
            toggleButton(v.Btn) : the button that activate the drawer
        """
        
        toggleButton.on_event('click', self._on_drawer_click)
            
        return self
    
    def _on_drawer_click(self, widget, event, data):
        """
        Toggle the drawer visibility
        """
        
        self.v_model = not self.v_model
        
        return self
        
    def _on_item_click(self, change):
        """
        Deactivate all the other items when on of the is activated
        """
        if change['new'] == False:
            return self
        
        # reset all others states
        for i in self.items:
            if i != change['owner']:
                i.input_value = False
            
        return self
        

class Footer(v.Footer, SepalWidget):
    """
    Custom Footer with cuzomizable text. 
    Not yet capable of displaying logos
    
    Args: 
        text (str, optional): the text to display in the future
    """
    def __init__(self, text="", **kwargs):
        
        text = text if text != '' else 'SEPAL \u00A9 {}'.format(datetime.today().year)
        
        super().__init__(
            color = sepal_main,
            class_ = "white--text",
            app=True,
            children = [text],
            **kwargs
        )
        
class App(v.App, SepalWidget):
        """
        Custom App display with the tiles created by the user using the sepal color framework.
        Display false appBar if not filled. Navdrawer is fully optionnal.
        The drawerItem will be linked to the app tile and they will be able to control their display
        If the navdrawer exist, it will be linked to the appbar togglebtn
        
        Args:
            tiles ([sw.Tile]): the tiles of the app
            appBar (sw.AppBar, optional): the appBar of the application
            footer (sw.Footer, optional): the footer of the application
            navDrawer (sw.NavDrawer): the navdrawer of the application
            
        Attributes:
            tiles ([sw.Tile]): the tiles of the app
            appBar (sw.AppBar, optional): the appBar of the application
            footer (sw.Footer, optional): the footer of the application
            navDrawer (sw.NavDrawer): the navdrawer of the application
            content (v.Content): the tiles organized in a fluid container
            
        """
        
        def __init__(self, tiles=[''], appBar=None, footer=None, navDrawer=None, **kwargs):
            
            self.tiles = None if tiles == [''] else tiles
            
            app_children = []
            
            # create a false appBar if necessary
            if not appBar:
                appBar = AppBar()
            self.appBar = appBar
            app_children.append(self.appBar)
            
            # add the navDrawer if existing
            self.navDrawer = None
            if navDrawer:
                # bind app tile list to the navdrawer
                for di in navDrawer.items:
                    di.display_tile(tiles)
                    
                # link it with the appbar
                navDrawer.display_drawer(self.appBar.toggle_button)
                
                # add the drawers to the children
                self.navDrawer = navDrawer
                app_children.append(self.navDrawer)

            # add the content of the app
            self.content = v.Content(children=[
                v.Container(fluid=True,children = tiles)
            ])
            app_children.append(self.content)
    
            # add the footer if existing
            if footer:
                self.footer = footer
                app_children.append(self.footer)
            
            super().__init__(
                v_model=None,
                children = app_children,
                **kwargs)
            
        def show_tile(self, name):
            """
            Select the tile to display when the app is launched
            
            Args:
                name (str): the mount-id of the tile(s) to display
            
            Return:
                self
            """
            # show the tile 
            for tile in self.tiles:
                if name == tile._metadata['mount_id']:
                    tile.show()
                else:
                    tile.hide()
            
            # activate the drawerItem
            if self.navDrawer:
                for i in self.navDrawer.items:
                    if name == i._metadata['card_id']:
                        i.input_value = True
                        
            return self


from markdown import markdown
from traitlets import Unicode
from pathlib import Path

import ipyvuetify as v

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget, Markdown
from sepal_ui.scripts import utils as su

class Tile(v.Layout, SepalWidget):
    """
    Custom Layout widget for the sepal UI framework. 
    It an helper to build a consistent tiling system. 
    Tile objects are indeed compatible with the other classes from sepal_ui.
    
    Args:
        id_ (str): the tile id that will be written in its mount_id _metadata attribute
        title (str): the title of the Tile
        inputs ([list]): the list of widget to display inside the tile
        btn (v.Btn): the process btn
        output (sw.Alert): the alert to display process informations to the end user
    """
    
    def __init__(self, id_, title, inputs=[''], btn=None, output=None, **kwargs):
        
        self.btn = btn
        if btn: inputs.append(btn)
        
        self.output = output
        if output: inputs.append(output)
        
        self.title = v.Html(xs12=True, tag='h2', children=[title])
        
        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]
        
        card = v.Card(
            class_ = "pa-5",
            raised = True,
            xs12 = True,
            children = [self.title] + content
        )
        
        super().__init__(
            _metadata={'mount_id': id_},
            row=True,
            align_center=True,
            class_="ma-5 d-inline",
            xs12=True,
            children = [card],
            **kwargs
        )
        
    def set_content(self, inputs):
        """
        Replace the current content of the tile with the provided inputs. it will keep the output and btn widget if existing.
        
        Args:
            inputs ([list]): the list of widget to display inside the tile
            
        Return:
            self
        """
        
        # create the widgets 
        content = [v.Flex(xs12=True, children=[widget]) for widget in inputs]
        
        # add the title  
        content = [self.children[0].children[0]] + content
        
        # add the output (if existing)
        if self.output:
            content = content + [self.output]
            
        if self.btn:
             content = content + [self.btn]
        
        self.children[0].children = content
        
        return self 
    
    def set_title(self, title):
        """
        Replace the current title
        
        Args:
            title (str): the new title of the object
            
        Return:
            self
        """
        
        self.title.children = [title]
        
        return self
    
    def get_title(self):
        """
        Return the current title of the tile
        
        Return:
            (str): the title
        """ 
        
        return self.title.children[0]
    
    def toggle_inputs(self, fields_2_show, fields):
        """
        Display only the widgets that are part of the input_list. the widget_list is the list of all the widgets of the tile.
    
        Args:
            fields_2_show ([v.widget]) : the list of input to be display
            fields ([v.widget]) : the list of the tile widget
            
        Return:
            self
        """
    
        for field in fields:
            if field in fields_2_show:
                su.show_component(field)
            else:
                su.hide_component(field)
                    
        return self
    
    def get_id(self):
        """
        Return the mount_id value
        
        Return:
            (str): the moun_id value from _metadata dict
        """
        
        return self._metadata['mount_id']
        
class TileAbout(Tile):
    """
    Create an about tile using a .md file. 
    This tile will have the "about_widget" id and "About" title.
    
    Args:
        pathname (str | pathlib.Path): the path to the .md file
    """
    
    def __init__(self, pathname, **kwargs):
        
        if type(pathname) == str:
            pathname = Path(pathname)
            
        #read the content and transform it into a html
        with pathname.open() as f:
            about = f.read()
        
        content = Markdown(about)
        
        super().__init__('about_widget', 'About', inputs=[content], **kwargs)
        
class TileDisclaimer(Tile):
    """
    Create a about tile using a the generic disclaimer .md file. 
    This tile will have the "about_widget" id and "Disclaimer" title.
    """
    
    def __init__(self, **kwargs):
        
        pathname = Path(__file__).parent.parent.joinpath('scripts', 'disclaimer.md')
        #pathname = os.path.join(os.path.dirname(__file__), '..', 'scripts', 'disclaimer.md')
        
        #read the content and transform it into a html
        with pathname.open() as f:
            disclaimer = f.read()
            
        content = Markdown(disclaimer)
        
        super().__init__('about_widget', 'Disclaimer', inputs=[content], **kwargs)


import ipyvuetify as v 

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from ..scripts import utils

class Btn(v.Btn, SepalWidget):
    """
    Custom process Btn filled with the provided text.
    the color will be defaulted to 'primary' and can be changed afterward according to your need
    
    Args:
        text (str, optional): the text to display in the btn
        icon (str, optional): the full name of any mdi-icon
        
    Attributes:
        v_icon (v.icon): the icon in the btn
    """
    
    def __init__(self, text='Click', icon=None, **kwargs):

        self.color='primary'
        self.v_icon = None
        self.children=[text]
        
        if icon:
            self.set_icon(icon)
        
        super().__init__(**kwargs)
        


    def set_icon(self, icon):
        """
        set a new icon
        
        Args:
            icon (str): the full name of a mdi-icon
            
        Return:
            self
        """
        if self.v_icon:
            self.v_icon.children = [icon]
        else:
            self.v_icon = v.Icon(left=True, children=[icon])
            self.children = [self.v_icon] + self.children
            
        return self
        
    def toggle_loading(self):
        """
        Jump between to states : disabled and loading - enabled and not loading
        
        Return:
            self
        """
        self.loading = not self.loading
        self.disabled = self.loading
        
        return self
    
class DownloadBtn(v.Btn, SepalWidget):
    """
    Custom download Btn filled with the provided text.
    the download icon is automatically embeded and green.
    The btn only accepts absolute links. if non is provided then the btn stays disabled
    
    Args:
        text (str): the message inside the btn
        path (str, optional): the absolute to a downloadable content    
    """
    
    def __init__(self, text, path='#', **kwargs):
        
        self.class_   = 'ma-2'
        self.xs5      = True
        self.color    = 'success'
        self.children = [
            v.Icon(left=True, children=['mdi-download']),
            text
        ]
        super().__init__(**kwargs)
        
        # create the url 
        self.set_url(path)
        
    def set_url(self, path='#'):
        """
        Set the url of the download btn. and unable it. 
        If nothing is provided the btn is disabled
        
        Args:
            path (str): the absolute path to a downloadable content
            
        Return:
            self
        """
        
        if utils.is_absolute(path):
            url = path
        else: 
            url = utils.create_download_link(path)
            
        self.href = url
        self.disabled = (path =='#')
        
        return self


from functools import partial
from pathlib import Path
import json
from datetime import datetime

import ipyvuetify as v
import geemap
import ee
import pandas as pd

from sepal_ui.aoi.aoi_io import Aoi_io
from sepal_ui import sepalwidgets as sw
from sepal_ui.scripts import utils as su, run_aoi_selection
from sepal_ui.mapping import SepalMap
from sepal_ui.message import ms
    
# initialize earth engine
su.init_ee()
    
class FileNameField(v.TextField, sw.SepalWidget):
    """
    custom v.TextField heriting from sw.SepalWidget. default_name will be used as default v_model
    
    Args:
        default_name (str): a default name
    """
    
    def __init__(self, default_name = None):
        
        super().__init__(
            label   = ms.aoi_sel.asset_name_lbl, 
            v_model = default_name
        )
        
class Adm0Select(v.Select, sw.SepalWidget):
    """ 
    Custom v.Select heriting from sw.SepalWidget. Uses the FAO GAUL 2015 admo0_name features as item
    """
    
    def __init__(self):
        country_codes = Path(__file__).parents[1].joinpath('scripts/country_code.csv')
        country_codes = pd.read_csv(country_codes)
        country_codes = country_codes.drop_duplicates(subset=['ADM0_CODE'])
        country_codes = country_codes.sort_values('ADM0_NAME')
        
        items = []
        for i, row in country_codes.iterrows():
            items.append({'text': row.ADM0_NAME, 'value': row.ADM0_CODE})
        
        country_codes = None
        
        super().__init__(
            items   = items, 
            label   = ms.aoi_sel.country_lbl, 
            v_model = None,
            dense = True
        )
        
class Adm1Select(v.Select, sw.SepalWidget):
    """ 
    Custom v.Select heriting from sw.SepalWidget. Uses the FAO GAUL 2015 admo1_name features as item
    """
    
    def __init__(self):
        
        super().__init__(
            items   = [], 
            label   = ms.aoi_sel.adm1_lbl, 
            v_model = None,
            dense = True
        )
        
    def set_admin_input(self, change):
        """set the list of possible items to the list of corresponding GAUL admo1_code - name"""
        
        adm0_code = change['new']
        
        # filter the country list to keep only each admin1 in the selected admin0
        country_codes = Path(__file__).parents[1].joinpath('scripts/country_code.csv')
        country_codes = pd.read_csv(country_codes)
        country_codes = country_codes[country_codes['ADM0_CODE'] == adm0_code]        
        country_codes = country_codes.drop_duplicates(subset=['ADM1_CODE'])
        country_codes = country_codes.sort_values('ADM1_NAME')
        
        # create a usable dict for the items
        items = []
        for i, row in country_codes.iterrows():
            items.append({'text': row.ADM1_NAME, 'value': row.ADM1_CODE})
        
        country_codes = None
        
        # set the items 
        self.items = items
        
        return self
        
        
class Adm2Select(v.Select, sw.SepalWidget):
    """ 
    Custom v.Select heriting from sw.SepalWidget. Uses the FAO GAUL 2015 admo0_name features as item
    """
    
    def __init__(self):
        
        super().__init__(
            items   = [], 
            label   = ms.aoi_sel.adm2_lbl, 
            v_model = None,
            dense = True
        )
        
    def set_admin_input(self, change):
        """set the list of possible items to the list of corresponding GAUL admo1_code - name"""
        
        adm1_code = change['new']
        
        # filter the country list to keep only each admin1 in the selected admin0
        country_codes = Path(__file__).parents[1].joinpath('scripts/country_code.csv')
        country_codes = pd.read_csv(country_codes)
        country_codes = country_codes[country_codes['ADM1_CODE'] == adm1_code]
        country_codes = country_codes.sort_values('ADM2_NAME')
        
        # create a usable dict for the items
        items = []
        for i, row in country_codes.iterrows():
            items.append({'text': row.ADM2_NAME, 'value': row.ADM2_CODE})
        
        country_codes = None
        
        # set the items 
        self.items = items
        
        return self
        
class MethodSelect(v.Select, sw.SepalWidget):
    """
    Custom v.Select heriting from sw.SepalWidget. create a method list based on the provided methods and default methods. 
    2 headers will be added (if relevant), on before the administrative selection methods, another before the custom asset/shape/points methods.
    if no list is provided the default_methods list will be used entirely.
    
    Args: 
        default_methods ([str]): the list of all methods name that we can to display
        methods ([str], optionnal): the list of all methods name that we want to display
         
    """
    
    def __init__(self, default_methods, methods=None):
        
        # get all the methods if none 
        methods = methods or default_methods
        
        # create a custom item list 
        items = []
        custom_header = True
        admin_header = True
        for m in methods:
            
            if m in default_methods:
                
                # admin header
                if m in default_methods[:3] and admin_header: 
                    items.append({'header': ms.aoi_sel.administrative})
                    admin_header = False
                elif m in default_methods[3:] and custom_header:
                    items.append({'header': ms.aoi_sel.custom})
                    custom_header = False
                        
                items.append({'text': m, 'value': m})
                
        # create the input 
        super().__init__(
            label = ms.aoi_sel.method_lbl,
            items = items,
            v_model = None,
            dense = True
        )
        
class TileAoi(sw.Tile):
    """
    sw.Tile tailored for the selection of an aoi. it is meant to be used with the aoi.Aoi_Io object.
    Render and bind all the variable to create an autonomous aoi selector. 
    If you use a custom aoi, it will create a asset in you gee account with the name 'aoi_[aoi_name]'.
    
    Args:
        io (aoi.Aoi_Io): an object that will carry the inputs and outputs of the tile
        methods ([str]): the methods to use for the aoi selection. needs to be part of the Private const SELECTION_METHOD
        folder (str | pathlib.path): the earthengine folder to use for asset saving and discovery. default to user Root
        
    Attributes: 
        io (aoi.aoi_Io): the aoi_io used to store aoi selection inputs and outputs
        
        aoi_output (sw.Alert): the Alert widget used to display information
        
        folder (str | pathlib.Path): the earthengine folder to use for asset saving and discovery. default to user Root
        
        aoi_file_name (FileNameField): a TextField input to choose asset name
        aoi_file_input (sw.FileInput): a file selector to retrieve sepal folders 
        aoi_country_selection (CountrySelect): the country selector input
        aoi_asset_name (sw.AssetSelect): a ComboBox input to select an asset in the self.folder or any custom asset link
        aoi_load_table (sw.LoadTableField): a table input field to retreive information from a point file
        aoi_select_method (v.Select): the input to select to aoi selection method to use. change the display of the tile widgets
        aoi_select_btn (sw.Btn): the Btn to launch the exportation of an asset
        
        m (ms.SepalMap): The map to display drawings and selected aoi. It is using both sattelites and CartoDb.DarkMatter basemaps
    """
    
    # constants
    SELECTION_METHOD =['Country boundaries', "First administrative layer", 'second administrative layer', 'Draw a shape', 'Use GEE asset', 'Upload file', 'Use points file']
    
    def __init__(self, io, methods = SELECTION_METHOD, folder = None, **kwargs):
        
        # load the io 
        self.io = io
        
        # create the output
        self.aoi_output = sw.Alert()
        
        # save the folder (mainly for testing purposes)
        self.folder = folder
        
        # create the inputs widgets 
        self.aoi_file_name = FileNameField(io.file_name).hide()     
        self.aoi_file_input = sw.FileInput(['.shp']).hide()
        self.aoi_country_selection = Adm0Select().hide()
        self.aoi_admin_1_select = Adm1Select().hide()
        self.aoi_admin_2_select = Adm2Select().hide()
        self.aoi_asset_name = sw.AssetSelect(folder = self.folder, default_asset = io.default_asset).hide()
        self.aoi_load_table = sw.LoadTableField().hide()
        
        # bind to aoi_io 
        self.aoi_output = sw.Alert() \
            .bind(self.aoi_file_name, self.io, 'file_name',  verbose=False) \
            .bind(self.aoi_file_input, self.io, 'file_input',  verbose=False) \
            .bind(self.aoi_country_selection, self.io, 'adm0',  verbose=False) \
            .bind(self.aoi_admin_1_select, self.io, 'adm1',  verbose=False) \
            .bind(self.aoi_admin_2_select, self.io, 'adm2',  verbose=False) \
            .bind(self.aoi_asset_name, self.io, 'assetId',  verbose=False) \
            .bind(self.aoi_load_table, self.io, 'json_csv',  verbose=False)
    
        widget_list = [
            self.aoi_file_name, 
            self.aoi_file_input, 
            self.aoi_country_selection, 
            self.aoi_admin_1_select,
            self.aoi_admin_2_select,
            self.aoi_asset_name,
            self.aoi_load_table
        ]
        
        #create the map 
        self.m = SepalMap(['Esri Satellite', 'CartoDB.DarkMatter'], dc=True)
    
        # bind the input to the selected method 
        #method_items = [m for m in methods if m in self.SELECTION_METHOD] 
        self.aoi_select_method = MethodSelect(self.SELECTION_METHOD, methods)
        
        # create the validation button 
        self.aoi_select_btn = sw.Btn(ms.aoi_sel.btn)
    
        # assemble everything on a tile 
        inputs = v.Layout(
            class_       = "pa-5",
            row          = True,
            align_center = True, 
            children     = (
                [v.Flex(xs12 = True, children =[self.aoi_select_method])] 
                + [v.Flex(xs12 = True, children = [widget]) for widget in widget_list]
                + [v.Flex(xs12 = True, children =[self.aoi_select_btn])] 
                + [v.Flex(xs12 = True, children = [self.aoi_output])]
            )
        )
        
        aoi_content_main = v.Layout(
            row      = True,
            xs12     = True,
            children = [
                v.Flex(xs12 = True, md6 = True, children = [inputs]),
                v.Flex(xs12 = True, md6 = True, class_ = "pa-5", children = [self.m])
            ]
        )
        
        super().__init__(id_ = 'aoi_widget', title = ms.aoi_sel.title, inputs = [aoi_content_main])
        
        # link the custom behaviours 
        self.aoi_load_table.observe(self._on_table_change, 'v_model')
        self.aoi_file_input.observe(self._on_file_change, 'v_model')
        self.m.dc.on_draw(self.handle_draw)
        self.aoi_select_method.observe(partial(self.bind_aoi_method, list_input = widget_list), 'v_model')
        self.aoi_select_btn.on_event('click', self.bind_aoi_process)
        self.aoi_country_selection.observe(self.aoi_admin_1_select.set_admin_input, 'v_model')
        self.aoi_admin_1_select.observe(self.aoi_admin_2_select.set_admin_input, 'v_model')
        
    def handle_draw(self, dc, action, geo_json):
        """
        handle the drawing of a geometry on a map. 
        The geometry is transform into a ee.featurecollection and save in self.io
        
        Args:
            dc (geemap.DrawingControl): the drawing control
            action (): the action that fire this handler
            geo_json (json): the dict of the drawn feature
            
        Return:
            self
        """
        
        # change the date 
        date = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.aoi_file_name.v_model = f'Manual_{date}'
        
        # add the feature
        geom = geemap.geojson_to_ee(geo_json, False)
        feature = ee.Feature(geom)
        self.io.drawn_feat = ee.FeatureCollection(feature)
            
        self.aoi_output.add_live_msg(ms.aoi_sel.shape_drawn)

        return self
    
    def bind_aoi_process(self, widget, event, data):
        """
        Use the user inputs to create a aoi. If a administrative layer have been selected, then its features are saved in self.io.
        If it's a custom selection method, a custom asset is created under `aoi_[aoi_name]` in the self.folder of the user GEE account.
        The final aoi will be displayed on the map 
        
        Args:
            widget (v.Vue): the vuetify object that trigger the binding
            event (str): the event that trigger the binding
            data ({str}): the data of the javascript event
            
        Return:
            self
        """
        
        # lock the btn
        widget.toggle_loading()            
            
        try:
            # create the aoi asset
            run_aoi_selection.run_aoi_selection(
                output      = self.aoi_output, 
                list_method = self.SELECTION_METHOD, 
                io          = self.io,
                folder      = self.folder
            )
            
            # display the resulting aoi on the map
            if self.io.get_aoi_ee():
                self.m.hide_dc()
                self.io.display_on_map(self.m)
            
        except Exception as e: 
            self.aoi_output.add_live_msg(str(e), 'error') 
        
        # free the btn
        widget.toggle_loading()
    
        return self
    
    def bind_aoi_method(self, change, list_input):
        """
        change the display of the AOI selector according to the method selected. will only display the useful inputs
        
        Args:
            change ({obj}): the change dictionnary of an observe method (see Traitlet documentation)
            list_input ([v.Vue]): the list of all the inputs of the aoi selector
            
        Return:
            self
        """
        
        # reset the aoi_io
        self.io.clear_attributes()
        
        # clearly identify the differents widgets 
        aoi_file_input        = list_input[1]
        aoi_file_name         = list_input[0]
        aoi_admin_0_select    = list_input[2]
        aoi_admin_1_select    = list_input[3]
        aoi_admin_2_select    = list_input[4]
        aoi_asset_name        = list_input[5]
        aoi_load_table        = list_input[6]
        
        # clear all the v_model
        aoi_file_name.v_model = None
        aoi_admin_0_select.v_model = None
        aoi_asset_name.v_model = None
        aoi_admin_1_select.items = []
        aoi_admin_2_select.items = []
        
        # extract the selecion methods
        method = self.SELECTION_METHOD
            
        # update the io
        self.io.selection_method = change['new']
            
        # hide the dc
        self.m.hide_dc()
        
        #############################################
        ##      toogle the appropriate inputs      ##
        #############################################
            
        # country selection
        if change['new'] == method[0]:
            self.toggle_inputs([aoi_admin_0_select], list_input)
        # admin 1
        elif change['new'] == method[1]: 
            self.toggle_inputs([aoi_admin_0_select, aoi_admin_1_select], list_input)
        # admin 2
        elif change['new'] == method[2]: 
            self.toggle_inputs([aoi_admin_0_select, aoi_admin_1_select, aoi_admin_2_select], list_input)
        # drawing
        elif change['new'] == method[3]: 
            self.toggle_inputs([aoi_file_name], list_input)
            self.m.show_dc()
        # gee asset
        elif change['new'] == method[4]: 
            self.toggle_inputs([aoi_asset_name], list_input)
        # shp file
        elif change['new'] == method[5]: 
            self.toggle_inputs([aoi_file_name, aoi_file_input], list_input)
        # Point file (.csv)
        elif change['new'] == method[6]: 
            self.toggle_inputs([aoi_file_name, aoi_load_table], list_input)
        # display nothing 
        else:
            self.toggle_inputs([], list_input)
        
        return self 
    
    def _on_file_change(self, change):
        """
        Change the aoi_file_name v_model with a pattern based on the file name
        the following pattern is used : 'aoi_[file_path.stem]'
        
        Args:
            change ({obj}): the change dictionnary of an observe method (see Traitlet documentation)
            
        Return:
            self
        """
        
        name = Path(change['new']).stem
        self.aoi_file_name.v_model = name
        
        return
    
    def _on_table_change(self, change):
        """
        Change the aoi_file_name v_model with a pattern based on the file name
        the following pattern is used : 'aoi_[file_path.stem]'
        
        Args:
            change ({obj}): the change dictionnary of an observe method (see Traitlet documentation)
            
        Return:
            self
        """
        
        load_df = json.loads(change['new'])
        
        name = Path(load_df['pathname']).stem
        self.aoi_file_name.v_model = name
        
        return
    


import time

import ee

from sepal_ui.message import ms
from sepal_ui.scripts import utils as su
    
# initialize earth engine
su.init_ee()

def wait_for_completion(task_descripsion, widget_alert=None):
    """
    Wait until the selected process is finished. Display some output information

    Args:
        task_descripsion (str): name of the running task
        widget_alert (v.Alert, optional): alert to display the output messages
        
    Return:
        (str): the final state of the task
    """
    state = 'UNSUBMITTED'
    while state != 'COMPLETED':
        
        # print in a widget
        if widget_alert: 
            widget_alert.add_live_msg(ms.status.format(state))
            
        # wait 5 seconds
        time.sleep(5)
                    
        # search for the task in task_list
        current_task = isTask(task_descripsion)
        state = current_task.state
        
        if state == 'FAILED':
            raise Exception(ms.status.format(state))
        
    # print in a widget
    if widget_alert: 
        widget_alert.add_live_msg(ms.status.format(state), 'success')
        
    return state
        
def isTask(task_descripsion):
    """
    Search for the described task in the user Task list return None if nothing is find
    
    Args: 
        task_descripsion (str): the task description
    
    Return:
        (ee.Task) : return the found task else None
    """
    
    current_task = None
    for task in ee.batch.Task.list():
        if task.config['description'] == task_descripsion:
            current_task = task
            break
            
    return current_task





from sepal_ui.sepalwidgets.sepalwidget import *
from sepal_ui.sepalwidgets.alert import *
from sepal_ui.sepalwidgets.btn import *
from sepal_ui.sepalwidgets.app import *
from sepal_ui.sepalwidgets.tile import *
from sepal_ui.sepalwidgets.inputs import *


from distutils.core import setup
    
setup(
    name = 'sepal_ui',      
    packages = ['sepal_ui', 'sepal_ui.scripts', 'sepal_ui.frontend', 'sepal_ui.sepalwidgets', 'sepal_ui.aoi', 'sepal_ui.message', 'sepal_ui.mapping', 'sepal_ui.translator'],   
    package_data={'sepal_ui': ['scripts/*.csv', 'scripts/*.md', 'message/*.json', 'bin/module_factory']},
    version = '1.1.5',   
    license='MIT',        
    description = 'Wrapper for ipyvuetify widgets to unify the display of voila dashboards in SEPAL platform',
    long_description = open('README.rst').read(),
    long_description_content_type = 'text/x-rst',
    author = 'Pierrick Rambaud',                   
    author_email = 'pierrick.rambaud49@gmail.com',  
    url = 'https://github.com/12rambau/sepal_ui',
    download_url = 'https://github.com/12rambau/sepal_ui/archive/v_1.1.5.tar.gz',
    keywords = ['UI', 'Python', 'widget', 'sepal'], 
    install_requires=[
        'haversine',
        'ipyvuetify',
        'geemap',
        'earthengine-api',
        'markdown',
        'xarray_leaflet',
        'shapely',
        'geopandas',
        'pandas',
        'deepdiff',
        'colorama',
        'Deprecated'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3.6',
    ],
)


import time
from datetime import datetime
import os
import glob
from pathlib import Path
import csv
from urllib.parse import urlparse
import subprocess
import string 
import random
import math
import base64
import os

import ipyvuetify as v
import pandas as pd
import ee
from cryptography.fernet import Fernet

import sepal_ui

def hide_component(widget):
    """
    hide a vuetify based component
    
    Args:
        widget (v.VuetifyWidget): the widget to hide
    """
    
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.hide()
    elif not 'd-none' in str(widget.class_):
        widget.class_ = str(widget.class_).strip() + ' d-none'
        
    return

def show_component(widget):
    """
    show a vuetify based component
    
    Args:
        widget (v.VuetifyWidget): the widget to hide
    """
    
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.show()
    elif 'd-none' in str(widget.class_):
        widget.class_ = widget.class_.replace('d-none', '')
        
    return 

def get_gaul_dic():
    """
    Create the list of the country code in the FAO GAUL norm using the CSV file provided in utils
        
    Return:
        (dict): the countries FAO_GAUL codes labelled with english country names
    """
    
    # file path
    path = Path(__file__).parent.joinpath('country_code.csv')
    
    # get the df and sort by country name
    df = pd.read_csv(path).sort_values(by=['ADM0_NAME'])
    
    # create the dict
    fao_gaul = {row['ADM0_NAME'] : row['ADM0_CODE'] for i, row in df.iterrows()}
        
    return fao_gaul

def get_iso_3(adm0):
    """
    Get the iso_3 code of a country_selection. Uses the fips_code if the iso-3 is not available
    
    Args:
        adm0 (int): the country adm0 code FAO GAUL 2015
        
    Return:
        (str): the 3 letters of the iso_3 country code
    """
    
    # file path
    path = Path(__file__).parent.joinpath('country_code.csv')
    
    # get the df
    df = pd.read_csv(path)
    
    row = df[df['ADM0_CODE'] == adm0]
    code = None
    if len(row):
        code = row['ISO 3166-1 alpha-3'].values[0]
        
    return code
    
def create_download_link(pathname):
    """
    Create a clickable link to download the pathname target
    
    Args:
        pathname (str | pathlib.Path): the pathname th download
        
    Return:
        (str): the download link
    """
    
    if type(pathname) == str:
        pathname = Path(pathname)
        
    result_path = Path(pathname).expanduser()
    home_path = Path('~').expanduser()
    
    # will be available with python 3.9
    #download_path = result_path.relative_to(home_path) if result_path.is_relative_to(home_path) else result_path
    download_path = os.path.relpath(result_path,home_path)
    
    link = f'/api/files/download?path=/{download_path}'
    
    return link

def is_absolute(url):
    """
    Check if the given url is an absolute or relative path
    
    Args:
        url (str): the url to test
        
    Return:
        (bool): True if absolute else False
    """
    return bool(urlparse(url).netloc)

def random_string(string_length=3):
    """
    Generates a random string of fixed length. 
    
    Args:
        string_length (int, optional): Fixed length. Defaults to 3.
    
    Return:
        (str): A random string
    """
    
    # random.seed(1001)
    letters = string.ascii_lowercase
    
    return ''.join(random.choice(letters) for i in range(string_length))

def get_file_size(filename):
    """
    Get the file size as string of 2 digit in the adapted scale (B, KB, MB....)
    
    Args:
        filename (str | pathlib.Path): the path to the file to mesure
        
    Return:
        (str): the file size in a readable humanly readable
    """
    
    file_size = Path(filename).stat().st_size
    
    if file_size == 0:
        return "0B"
    
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    
    i = int(math.floor(math.log(file_size, 1024)))
    s = file_size / (1024 ** i)
        
    return '{:.1f} {}'.format(s, size_name[i])

def init_ee():
    """
    Initialize earth engine according to the environment. 
    It will use the creddential file if the EE_PRIVATE_KEY env variable exist. 
    Otherwise it use the simple Initilize command (asking the user to register if necessary)
    """
    
    # only do the initialization if the credential are missing
    if not ee.data._credentials:
        
        # if the decrypt key is available use the decript key 
        if 'EE_DECRYPT_KEY' in os.environ:
        
            # read the key as byte 
            key = os.environ['EE_DECRYPT_KEY'].encode()
            
            # create the fernet object 
            fernet = Fernet(key)
            
            # decrypt the key
            json_encrypted = Path(__file__).parent.joinpath('encrypted_key.json')
            with json_encrypted.open('rb') as f:
                json_decripted = fernet.decrypt(f.read()).decode()
                
            # write it to a file
            with open('ee_private_key.json', 'w') as f:
                f.write(json_decripted)
                
            # connection to the service account
            service_account = 'test-sepal-ui@sepal-ui.iam.gserviceaccount.com'
            credentials = ee.ServiceAccountCredentials(service_account, 'ee_private_key.json')
            ee.Initialize(credentials)
        
        # if in local env use the local user credential
        else:
            ee.Initialize()
            
    return


from pathlib import Path

import ee
import geemap
import shapely.geometry as sg
import geopandas as gpd
import ipyvuetify as v 

from sepal_ui.scripts import utils as su
    
# initialize earth engine
su.init_ee()

class Aoi_io:
    """
    an io object dedicated to the sorage and the manipulation of aoi. 
    It is meant to be used with the TileAoi object. 
    By using this you will be able to provide your application with aoi as an ee_object or any other intersting format.
    The class also provide insight on your aoi.
    
    Args: 
        alert_widget (sw.Alert): an alert output to display message when computing stuff internally. This is a legacy parameter
        default_asset (str): the default asset to use when no others are provided
        default_admin0 (str): a default administrative layer of level 0 to use if nothing is provided (will overload the default_asset). use a FAO gaul 2015 id
        
    Attributes:
        default_asset (str): the default asset link
        default_admin0 (int): the default country Id (in FAO GAUL 2015)
        
        assetId (str): the assetId of the current asset
        column (str): name of a aoi's column (feature)  
        field (str): name or value of a specific filed in a specific column
        selected_feature (ee.Feature): a feature queried inside the ee.FeatureCollection version of the object
        json_csv (str | pathlib.Path): the path to the json initial file
        adm0 (int): the code of the admin0 level in FAO GAUL 2015 format
        adm1 (int): the code of the admin1 level in FAO GAUL 2015 format
        adm2 (int): the code of the admin2 level in FAO GAUL 2015 format
        feature_collection (ee.FeatureCollection): the feature_collection to use as asset when dealing with administrative layers
        
        file_input (str | pathlib.Path): the path to the input .shp file
        file_name (str): the name that will be used to name the AOI when exporting
        country_selection (str): the name of the country being used 
        selection_method (str): the selection method used in the linked aoi.TileAoi
        drawn_feat (geo_json): the feature drawn on the map
        alert (sw.Alert): the alert to display message
    """
    
    def __init__(self, alert_widget=None, default_asset=None, default_admin0=None, default_admin1=None, default_admin2=None):
        
        # keep the default assets in memory
        self.default_asset = default_asset
        self.default_admin0 = default_admin0
        self.default_admin1 = default_admin1
        self.default_admin2 = default_admin2
        
        # GEE parameters
        self.assetId = self.default_asset
        self.column = None
        self.field = None
        self.selected_feature = None
        self.json_csv = None # information that will be use to transform the csv into asset 
        self.adm0 = self.default_admin0 
        self.adm1 = self.default_admin1
        self.adm2 = self.default_admin2 
        self.feature_collection = None # to access the country asset
        
        # set the feature_collection 
        self._set_default_admin()

        #set up your inputs
        self.file_input = None
        self.file_name = None
        self.selection_method = None
        self.drawn_feat = None
        self.alert = alert_widget
        
    def is_admin(self):
        """
        Test if the current aoi_io object is refeering to an administrative layer or not
        
        Return:
            (bool): True if administrative layer else False. False as well if no aoi is selected.
        """
        admin = False
        
        if self.adm0 or self.adm1 or self.adm2:
            admin = True
            
        return admin
        
    def get_aoi_ee(self):
        """ 
        get the ee_object corresponding to current self, None if no aoi set.
            
        Return:
            (ee.FeatureCollection): the aoi FeatureCollection in GEE format
        """
        
        obj = None
        if self.is_admin():
            obj = self.feature_collection
        elif self.assetId:
            obj = ee.FeatureCollection(self.assetId)
            
        return obj
    
    def get_columns(self):
        """ 
        Retrieve the columns or variables from self excluding `system:index` and `Shape_Area`.

            Return: 
                ([str]): sorted list cof column names
        """
        
        aoi_ee = self.get_aoi_ee()
        columns = ee.Feature(aoi_ee.first()).propertyNames().getInfo()
        columns = sorted([col for col in columns if col not in ['system:index', 'Shape_Area']])
        
        return columns
    
    def get_fields(self, column=None):
        """" 
        Retrieve the fields from a column. It will use the self.column if not provided.
        
        Args:
            column (str, optional): A column name to query over the asset
        
        Return: 
            ([str]): sorted list of fields value

        """
        
        if not column:
            column = self.column

        aoi_ee = self.get_aoi_ee()
        fields = sorted(aoi_ee.distinct(column).aggregate_array(column).getInfo())

        return fields

    def get_selected_feature(self):
        """ 
        Select an ee object based on the current `self.column` and `self.field`.

        Return:
            (ee.Geometry): the geometry associated with the query
        """

        if not self.column or not self.field:
            raise Exception('You must first select a column and a field.')

        ee_asset = self.get_aoi_ee()
        select_feature = ee_asset.filterMetadata(self.column, 'equals', self.field).geometry()

        # Specify the selected feature
        self.selected_feature = select_feature

        return select_feature

    def clear_selected(self):
        """
        clear the selected_feature attribute.
        
        Return:
            self
        """
        self.selected_feature = None
        
        return self

    def clear_attributes(self):
        """
        Return all attributes to their default state.
        Set the default_asset as current assetId.
        
        Return: 
            self
        """

        # GEE parameters
        self.assetId = self.default_asset
        self.column = None
        self.field = None
        self.selected_feature = None
        self.json_csv = None
        self.adm0 = self.default_admin0
        self.adm1 = self.default_admin1
        self.adm2 = self.default_admin2
        self.feature_collection = None
        
        # set the feature_collection
        self._set_default_admin()

        # set up your inputs
        self.file_input = None
        self.file_name = None
        self.selection_method = None
        self.drawn_feat = None

        return self
    
    def get_not_null_attrs(self):
        """
        Retrieve all the non null attributes of the object.
        
        Return:
            ([str]): The list of the non null attributes' name
        """
        attrs = dict((k, v) for k, v in self.__dict__.items() if v is not None)
        
        return attrs

    def display_on_map(self, map_):
        """ 
        Display the current aoi on a `ms.SepalMap`.
        The drawing control of the map will be removed if existing.

        Args:
            map_ (ms.SepalMap): Map to display the element
            
        Return:
            self
        """
        aoi = self.get_aoi_ee()
        map_.zoom_ee_object(aoi.geometry())
        map_.addLayer(aoi, {'color': v.theme.themes.dark.success}, name='aoi')
        
        map_.hide_dc()
        
        return self


    def get_bounds(self, ee_object, cardinal=False):
        """ 
        Get the min(lon,lat) and max(lon, lat) from the given `ee` object.
        returns coordinates (lon, lat) of each cardinal points (tl, bl, tr, br) if cardinal is `True` else returns a bounding box (min_lon, min_lat, max_lon, max_lat)

        Args:
            ee_asset (ee.object): GEE asset (FeatureCollection, Geometry or str)
            cardinal (boolean) (optional)

        Return:
            (tuple(tuple(int)) | tuple(int)): coordinates (lon, lat) of each cardinal points or a bounding box
        """ 
            
        ee_bounds = ee.FeatureCollection(ee_object).geometry().bounds().coordinates()
        coords = ee_bounds.get(0).getInfo()
        ll, ur = coords[0], coords[2]

        # Get the bounding box
        min_lon, min_lat, max_lon, max_lat = ll[0], ll[1], ur[0], ur[1]

        # Get (x, y) of the 4 cardinal points
        tl = (min_lon, max_lat)
        bl = (min_lon, min_lat)
        tr = (max_lon, max_lat)
        br = (max_lon, min_lat)

        return (tl, bl, tr, br) if cardinal else (min_lon, min_lat, max_lon, max_lat)
    
    def get_aoi_shp(self, dwnDir=''):
        """ 
        Create the .shp file corresponding to the selected aoi
        
        Args:
            dwnDir (str | pathlib.Path): the path to directory that will store the ESRI shapefiles
            
        Return:
            (str): the absolute path to the .shp file
        """
        
        aoi_name = self.get_aoi_name()
        
        if type(dwnDir) == str:
            dwnDir = Path(dwnDir).expanduser()
            
        filename = dwnDir.joinpath(f'{aoi_name}.shp')
    
        if filename.is_file():
            return filename
    
        # verify that the asset exist
        aoi = self.get_aoi_ee()
    
        # convert into json
        aoi_json = geemap.ee_to_geojson(aoi)
        
        # convert to geopandas gdf
        gdf = gpd.GeoDataFrame.from_features(aoi_json).set_crs('EPSG:4326')
        
        # FAO GAUL geometries are full of linestring which cannot be converted into shapefile so some filtering needs to be done first 
        gdf_filtered = gdf.copy()

        for i, row in gdf.iterrows():
            if type(row.geometry) == sg.collection.GeometryCollection:

                # get the polygon and only keep the polygon 
                for shape in row.geometry:
                    if type(shape) == sg.polygon.Polygon:
                        gdf_filtered.at[i, 'geometry'] = shape
                        break
        
        gdf_filtered.to_file(filename)
    
        return filename
    
    def get_aoi_name(self):
        """ 
        Get the aoi name based on the current state. 
        It will be an ISO 3166-1 alpha-3 if it's a country name and the asset stem for the rest. 
        Remove the `aoi_` string before the name of asset as it's a creation prefix convention.
        Return None if there is nothing. 
        
        Return:
            (str): the aoi name
        """
        
        name = None
        if self.is_admin():
            name = su.get_iso_3(self.adm0)
            if self.adm1:
                name += f"_{self.adm1}"
            if self.adm2:
                name += f"_{self.adm2}"
        elif self.assetId:
            name = Path(self.assetId).stem.replace('aoi_', '')
        
        return name
    
    def _set_default_admin(self):
        """
        Set the self.feature_collection acording to the default admin code.
        It will return an empty featurecollection if the value is not reffering to anything
        
        Return:
            self
        """
        
        if self.default_admin0:
            self.feature_collection = ee.FeatureCollection("FAO/GAUL/2015/level0").filter(ee.Filter.eq('ADM0_CODE', self.default_admin0))
            
        return self
    
    def set_asset(self, assetId):
        """
        Set the asset id of the aoi and unsure that all existing reference to administrative layer are removed
        
        Args:
            assetId (str): the assetId of the new asset
            
        Return:
            self
        """
        
        # add the asset 
        self.assetId = assetId 
        
        # remove all the administrative layers 
        self.feature_collection = None
        self.adm0 = None
        self.adm1 = None
        self.adm2 = None
        
        return self 
    
    def set_admin(self, asset, admin0=None, admin1=None, admin2=None):
        """
        Set the asset in feature_collection and referenced the admin level code
        Remove all existing assetId. At least one of the admin level need to be filled
        
        Args:
            admin0 (int, optional): the adm0 code in the FAO GAUL 2015 code list
            admin1 (int, optional): the adm1 code in the FAO GAUL 2015 code list
            admin2 (int, optional): the adm2 code in the FAO GAUL 2015 code list
            asset (ee.FeatureCollection): the asset associated with the code
            
        Return: 
            self
        """
        
        # add the asset 
        self.feature_collection = asset 
        
        # remove the assetId 
        self.assetId = None
        
        # add the admin level code 
        if (admin0, admin1, admin2) == (None, None, None):
            raise Exception("Impossible to set an administrative level without level")
        
        self.adm0 = admin0
        self.adm1 = admin1
        self.adm2 = admin2
            
        return self
            


# knwon bug of rasterio
import os 
if 'GDAL_DATA' in list(os.environ.keys()): del os.environ['GDAL_DATA']
if 'PROJ_LIB' in list(os.environ.keys()): del os.environ['PROJ_LIB']

import collections
from pathlib import Path

import ee 
import geemap
from haversine import haversine
import xarray_leaflet
import numpy as np
import rioxarray
import xarray as xr
import matplotlib.pyplot as plt
import ipywidgets as widgets
from ipyleaflet import (
    AttributionControl, DrawControl, LayersControl, 
    LocalTileLayer, ScaleControl, TileLayer, WidgetControl,
    ZoomControl
)
from traitlets import (
    Bool, link, observe
)
import ipyvuetify as v
from deprecated import deprecated

from sepal_ui.scripts import utils as su
from sepal_ui.message import ms

#initialize earth engine
su.init_ee()

class SepalMap(geemap.Map):
    """
    The SepalMap class inherits from geemap.Map. It can thus be initialized with all its parameter. 
    The map will fall back to CartoDB.DarkMatter map that well fits with the rest of the sepal_ui layout.
    Numerous methods have been added in the class to help you deal with your workflow implementation.
    It can natively display raster from .tif files and files and ee objects using methods that have the same signature as the GEE JavaScripts console
    
    Args: 
        basemaps ['str']: the basemaps used as background in the map. If multiple selection, they will be displayed as layers.
        dc (bool): wether or not the drawing control should be displayed
        vinspector (bool) : Add value inspector to map, useful to inspect pixel values
        
    Attributes:
        loaded_rasters ({geemap.Layer}): the raster that are already loaded in the map
        output_r (ipywidgets.Output): the rectangle to display the result of the raster interaction
        output_control_r (ipyleaflet.WidgetControl): the custom control on the map
        dc (ipyleaflet.DrawingControl): the drawing control of the map 
        
    """

    vinspector = Bool(False).tag(sync=True)
    
    def __init__(self, basemaps=[], dc=False, vinspector=False, **kwargs):
        
        # Initial parameters


        super().__init__(
            add_google_map=False,
            center = [0,0],
            zoom = 2,
            **kwargs
        )
        
        # init the rasters
        self.loaded_rasters = {}
        
        # add the basemaps
        self.clear_layers()
        if len(basemaps):
            for basemap in set(basemaps):
                self.add_basemap(basemap)
        else:
            self.add_basemap('CartoDB.DarkMatter')
        
        # add the base controls
        self.clear_controls()
        self.add_control(ZoomControl(position='topright'))
        self.add_control(LayersControl(position='topright'))
        self.add_control(AttributionControl(position='bottomleft'))
        self.add_control(ScaleControl(position='bottomleft', imperial=False))
        
        # specific drawing control
        self.set_drawing_controls(dc)
        
        # Add value inspector
        self.w_vinspector = widgets.Checkbox(
                    value=False,
                    description='Inspect values',
                    indent=False,
                    layout=widgets.Layout(width='18ex')
        )

        if vinspector:
            self.add_control(
                WidgetControl(
                    widget = self.w_vinspector,
                    position = 'topright')
            )
            link((self.w_vinspector, 'value'),(self, 'vinspector'))

        # Create output space for raster interaction
        self.output_r = widgets.Output(layout={'border': '1px solid black'})
        self.output_control_r = WidgetControl(widget=self.output_r, position='bottomright')
        self.add_control(self.output_control_r)

        # define interaction with rasters
        self.on_interaction(self.raster_interaction)
        
    @observe('vinspector')
    def change_cursor(self, change):
        """Method to be called when vinspector trait changes """
        
        if self.vinspector:
            self.default_style = {'cursor': 'crosshair'}
        else:
            self.default_style = {'cursor': 'grab'}
            
        return
        
    def raster_interaction(self, **kwargs):
        """Define a behavior when ispector checked and map clicked"""
        
        if kwargs.get('type') == 'click' and self.vinspector:
            latlon = kwargs.get('coordinates')
            self.default_style = {'cursor': 'wait'}

            local_rasters = [lr.name for lr in self.layers if isinstance(lr, LocalTileLayer)]

            if local_rasters:

                with self.output_r: 
                    self.output_r.clear_output(wait=True)
                    
                    for lr_name in local_rasters:

                        lr = self.loaded_rasters[lr_name]
                        lat, lon = latlon

                        # Verify if the selected latlon is the image bounds
                        if any([lat<lr.bottom, lat>lr.top, lon<lr.left, lon>lr.right]):
                            print('Location out of raster bounds')
                        else:
                            # row in pixel coordinates
                            y = int(((lr.top - lat) / abs(lr.y_res)))

                            # column in pixel coordinates
                            x = int(((lon - lr.left) / abs(lr.x_res)))

                            # get height and width
                            h, w = lr.data.shape
                            value = lr.data[y][x]
                            print(f'{lr_name}')
                            print(f'Lat: {round(lat,4)}, Lon: {round(lon,4)}')
                            print(f'x:{x}, y:{y}')
                            print(f'Pixel value: {value}')
            else:
                with self.output_r:
                    self.output_r.clear_output()

            self.default_style = {'cursor': 'crosshair'}
            
            return

    def set_drawing_controls(self, add=False):
        """
        Create a drawing control for the map.
        It will be possible to draw rectangles, circles and polygons.
        
        Args:
            add (bool): either to add the dc to the object attribute or not
            
        return:
            self
        """
        
        color = v.theme.themes.dark.info
        
        dc = DrawControl(
            marker       = {},
            circlemarker = {},
            polyline     = {},
            rectangle    = {'shapeOptions': {'color': color}},
            circle       = {'shapeOptions': {'color': color}},
            polygon      = {'shapeOptions': {'color': color}},
        )
            
        self.dc = dc if add else None
        
        return self

    def _remove_local_raster(self, local_layer):
        """ 
        Remove local layer from memory
        
        Args:
            local_layer (str | geemap.layer): The local layer to remove or its name
            
        Return: 
            self
        """
        name = local_layer if type(local_layer) == str else local_layer.name
        
        if name in self.loaded_rasters.keys():
            self.loaded_rasters.pop(name)
            
        return self

    def remove_last_layer(self, local=False):
        """
        Remove last added layer from Map

        Args:
            local (boolean): Specify True to only remove local last layers, otherwise will remove every last layer.

        Return:
            self
        """
        if len(self.layers) > 1:

            last_layer = self.layers[-1]

            if local:
                local_rasters = [lr for lr in self.layers if isinstance(lr, LocalTileLayer)]
                if local_rasters:
                    last_layer = local_rasters[-1]
                    self.remove_layer(last_layer)

                    # If last layer is local_layer, remove it from memory
                    if isinstance(last_layer, LocalTileLayer):
                        self._remove_local_raster(last_layer)
            else:
                self.remove_layer(last_layer)

                # If last layer is local_layer, remove it from memory
                if isinstance(last_layer, LocalTileLayer):
                    self._remove_local_raster(last_layer)
                    
        return self
              
    def zoom_ee_object(self, ee_geometry, zoom_out=1):
        """ 
        Get the proper zoom to the given ee geometry.

        Args:
            ee_geometry (ee.Geometry): the geometry to zzom on
            zoom_out (int) (optional): Zoom out the bounding zoom
            
        Return:
            self
        """
        
        # center the image
        self.centerObject(ee_geometry)
        
        # extract bounds from ee_object 
        ee_bounds = ee_geometry.bounds().coordinates()
        coords = ee_bounds.get(0).getInfo()
        
        # Get (x, y) of the 4 cardinal points
        bl, br, tr, tl, _ = coords
        
        # Get (x, y) of the 4 cardinal points
        min_lon, min_lat = bl
        max_lon, max_lat = tr
        
        #zoom on these bounds 
        self.zoom_bounds([tl, bl, tr, br], zoom_out)
        
        return self 
    
    def zoom_bounds(self, bounds, zoom_out=1):
        """ 
        Adapt the zoom to the given bounds.

        Args:
            bounds (list of tuple(x,y)): coordinates of tl, bl, tr, br points
            zoom_out (int) (optional): Zoom out the bounding zoom
            
        Return:
            self
        """
        
        tl, bl, tr, br = bounds        
        
        maxsize = max(haversine(tl, br), haversine(bl, tr))
        
        lg = 40075 # number of displayed km at zoom 1
        zoom = 1
        while lg > maxsize:
            zoom += 1
            lg /= 2

        if zoom_out > zoom:
            zoom_out = zoom - 1

        self.zoom = zoom-zoom_out
        
        return self
    
    @deprecated(reason="will be removed in version 2.0")
    def update_map(self, assetId, bounds, remove_last=False):
        """
        Update the map with the asset overlay
        
        Args:
            assetId (str): the asset ID in gee assets
            bounds (list of tuple(x,y)): coordinates of tl, bl, tr, br points
            remove_last (boolean) (optional): Remove the last layer (if there is one) before updating the map
        """  
        if remove_last:
            self.remove_last_layer()

        self.zoom_bounds(bounds, zoom_out=2)
        self.centerObject(ee.FeatureCollection(assetId), zoom=self.zoom)
        self.addLayer(ee.FeatureCollection(assetId), {'color': 'green'}, name='aoi')
        
        return self
    
    # copy of the geemap add_raster function to prevent a bug from sepal 
    def add_raster(
        self, 
        image, 
        bands=None, 
        layer_name='Layer_' + su.random_string(), 
        colormap=plt.cm.inferno, 
        x_dim='x', 
        y_dim='y', 
        opacity=1.0, 
        fit_bounds=True,
        get_base_url=lambda _: 'https://sepal.io/api/sandbox/jupyter',
        colorbar_position='bottomright'):
        """
        Adds a local raster dataset to the map.
        
        Args:
            image (str | pathlib.Path): The image file path.
            bands (int or list, optional): The image bands to use. It can be either a number (e.g., 1) or a list (e.g., [3, 2, 1]). Defaults to None.
            layer_name (str, optional): The layer name to use for the raster. Defaults to None.
            colormap (str, optional): The name of the colormap to use for the raster, such as 'gray' and 'terrain'. More can be found at https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html. Defaults to None.
            x_dim (str, optional): The x dimension. Defaults to 'x'.
            y_dim (str, optional): The y dimension. Defaults to 'y'.
            fit_bounds (bool, optional): Wether or not we should fit the map to the image bounds. Default to True.
            get_base_url (callable, optional): A function taking the window URL and returning the base URL to use. It's design to work in the SEPAL environment, you only need to change it if you want to work outside of our platform. See xarray-leaflet lib for more details.
            colorbar_position (str, optional): The position of the colorbar (default to "bottomright"). set to False to remove it. 
        """
        
        if type(image) == str:
            image = Path(image)
            
        if not image.is_file():
            raise Exception(ms.mapping.no_image)
            
        # check inputs
        if layer_name in self.loaded_rasters.keys():
            layer_name = layer_name+su.random_string()

        if isinstance(colormap, str):
            colormap = plt.cm.get_cmap(name=colormap)

        da = rioxarray.open_rasterio(image, masked=True)

        # Create a named tuple with raster bounds and resolution
        local_raster = collections.namedtuple(
            'LocalRaster', ('name', 'left', 'bottom', 'right', 'top', 'x_res', 'y_res', 'data')
            )(layer_name, *da.rio.bounds(), *da.rio.resolution(), da.data[0])

        self.loaded_rasters[layer_name] = local_raster


        multi_band = False
        if len(da.band) > 1:
            multi_band = True
            if not bands:
                bands = [3, 2, 1]
        else:
            bands = 1

        if multi_band:
            da = da.rio.write_nodata(0)
        else:
            da = da.rio.write_nodata(np.nan)
        da = da.sel(band=bands)

        kwargs = {
            'm': self,
            'x_dim': x_dim,
            'y_dim': y_dim,
            'fit_bounds': fit_bounds,
            'get_base_url': get_base_url,
            #'colorbar_position': colorbar_position # will be uncoment when the colobared version of xarray-leaflet will be released
            'rgb_dim': 'band' if multi_band else None,
            'colormap': None if multi_band else colormap,
        }
        
        # display the layer on the map
        layer = da.leaflet.plot(**kwargs)

        layer.name = layer_name

        
        layer.opacity = opacity if abs(opacity) <= 1.0 else 1.0
        
        return
    
    def show_dc(self):
        """
        show the drawing control on the map
        
        Return:
            self
        """
        
        if self.dc:
            self.dc.clear()
        
            if not self.dc in self.controls:
                self.add_control(self.dc)
            
        return self
    
    def hide_dc(self):
        """
        hide the drawing control of the map
        
        Return:
            self
        """
        
        if self.dc:
            self.dc.clear()
        
            if self.dc in self.controls:
                self.remove_control(self.dc)
                
        return self
        
    @staticmethod   
    def get_basemap_list():
        """
        This function is intending for development use
        It give the list of all the available basemaps for SepalMap object
        
        Return:
            ([str]): the list of the basemap names
        """
        
        return [k for k in geemap.ee_basemaps.keys()]
        
        



import os
from pathlib import Path

from sepal_ui.translator import Translator

# pick the language
lang = 'en'
if 'CUSTOM_LANGUAGE' in os.environ:
    lang = os.environ['CUSTOM_LANGUAGE']

ms = Translator(Path(__file__).parent, lang)


import json
from types import SimpleNamespace
from pathlib import Path
from collections import abc
from deepdiff import DeepDiff


class Translator(SimpleNamespace):
    """
    The translator is a SimpleNamespace of Simplenamespace. It read 2 json file, the first one being the source language (usually english) and the second one the target language. 
    It will replace in the source dict every key that exist in both json dict. Following this procedure, every message that is not translated can still be accessed in the source language. 
    To access the dict keys, instead of using [], you can simply use key name as in an object ex: translator.first_key.secondary_key. 
    There are no depth limits, just respect the snake_case convention when naming your keys in the .json dicts.
    
    Args: 
        json_folder (str | pathlib.Path): the folder where the dictionnaries are stored
        target_lan (str): the language code of the target lang (it should be the same as the target dictionnary)
        default_lan (str): the language code of the source lang (it should be the same as the source dictionnary)
        
    Attributes:
        default_dict (dict): the source language dictionnary
        target_dict (dict): the target language dictionnary 
        (keys): all the keys can be acceced as attributes. make sure to never use default_dict and target_dict
    
    """
    def __init__(self, json_folder, target_lan, default_lan='en'):
        
        super().__init__()
        
        if type(json_folder) == str:
            json_folder = Path(json_folder)
        
        # reading both the default dict
        self.default_dict = json.loads(json_folder.joinpath(f'{default_lan}.json').read_text())
        
        # create a composite dict replaceing all the default keys with the one availabel in the target lan        
        target_path = json_folder.joinpath(f'{target_lan}.json')
        self.target_dict = self.default_dict.copy()
        
        if target_path.is_file():
            self.target_dict = json.loads(target_path.read_text())
        else:
            print(f'No json file is provided for "{target_lan}", fallback to "en"')

        
        # create the composite dictionnary
        ms_dict = self._update(self.default_dict, self.target_dict)
        
        # verify if 'default_dict' or 'target_dict' is in use
        self.search_key(ms_dict, 'default_dict')
        self.search_key(ms_dict, 'target_dict')
        
        # transform it into a json str
        ms_json = json.dumps(ms_dict)
        
        # unpack the json as a simple namespace
        ms = json.loads(ms_json, object_hook=lambda d: SimpleNamespace(**d))
        
        for k, v in ms.__dict__.items():
            setattr(self, k, getattr(ms, k))
    
    @classmethod
    def search_key(cls, d, key):
        """
        Search a specific key in the d dictionnary and raise an error if found
        
        Args:
            d (dict): the dictionnary to study 
            key (str): the key to look for
        """
        
        for k, v in d.items():
            if isinstance(v, abc.Mapping):
                cls.search_key(v, key)
            else:
                if k == key:
                    raise Exception(f"You cannot use the key {key} in your translation dictionnary")
                    break
        
        return None
        
    def _update(self, d, u):
        """ 
        Update the fallback dictionnaire (d) values with the keys that exist in the target (u) dictionnaire
        
        Args:
            d (dict): The fallback dictionnary
            u (dict): the target dctionnary
            
        Return:
            ms (dict): The updated dictionnay
        """
        
        ms = d.copy()
        
        for k, v in u.items():
            if isinstance(v, abc.Mapping):
                ms[k] = self._update(d.get(k, {}), v)
            else:
                ms[k] = v
        
        return ms
    
    def missing_keys(self):
        """
        this function is intended for developer use only
        print the list of the missing keys in the target dictionnairie
        
        Return:
            (str): the list of missing keys
        """
        
        # find all the missing keys
        try: 
            ddiff = DeepDiff(self.default_dict, self.target_dict)['dictionary_item_removed']
        except:
            ddiff = ['All messages are translated']
        
        return  '\n'.join(ddiff)
        
        